#include "checksum.h"
#include "map.h"
#include "fileio.h"


//////// Simple Transfer checksum ////////
// Assembly rips by Coconut Emulator

BYTE simpleChecksum(void * buffer, Uint32 len)
{
	BYTE c = 0;

	for (Uint32 i = 0; i < len; i++)
	{
		c ^= ((BYTE*)buffer)[i];
	}

	return c;
}


//////// File checksum ////////
// This section made possible by Snrrrub

Uint32 getFileChecksum(BYTE * buffer, long * dictionary, Uint32 len)
{
	Uint32 Index = 0,
		   Key   = Index - 1;

	for (Uint32 i = 0; i < len; i++)
	{
		Index	= dictionary[(Key & 255) ^ buffer[i]];
		Key		= (Key >> 8) ^ Index;
	}

	return ~Key;
}

Uint32 getFileChecksum(char * fileName)
{
	Uint32 chk = -1;

	fileAccess fa;
	if (fa.open(fileName, true))
	{
		Uint32 len = fa.getLength();
		BYTE * file = (BYTE*)fa.readStuff(0, len);
		long dictionary[256];
		generateDictionary(dictionary, 0);
		chk = getFileChecksum(file, dictionary, len);
		delete []file;
	}

	return chk;
}

void generate4(long * offset, long key)
{
	offset[0] = key;
	offset[1] = key ^ G4_MODIFIER;
	offset[2] = key ^= (G4_MODIFIER << 1);
	offset[3] = key ^ G4_MODIFIER;
}

void generate16(long * offset, long key)
{
	generate4(offset, key);
	generate4(offset + 4, key  ^ G16_MODIFIER);
	generate4(offset + 8, key  ^= (G16_MODIFIER << 1));
	generate4(offset + 12, key ^ G16_MODIFIER);
}

void generate64(long * offset, long key)
{
	generate16(offset, key);
	generate16(offset + 16, key ^ G64_MODIFIER);
	generate16(offset + 32, key ^= (G64_MODIFIER << 1));
	generate16(offset + 48, key ^ G64_MODIFIER);
}

void generateDictionary(long * offset, long key)
{
	generate64(offset, key);
	generate64(offset + 64, key  ^ G256_MODIFIER);
	generate64(offset + 128, key ^= (G256_MODIFIER << 1));
	generate64(offset + 192, key ^ G256_MODIFIER);
}


//////// Security checksum ////////
//  Supreme thanks go to Coconut Emulator and Snrrrub
//  for providing the assembly rips.

Uint32 generateLevelChecksum(Uint32 key, char * mapData)
{
	// Generate initial y
	Uint32 y = key;

	if (key & 0x80000000)
	{	// key is negative
		y = ~y + 1;								// make it positive

		y &= 31;								// take it modulo 32

		y = ~y + 1;								// make it negative

		if ((long)y > 1023) return key;			// must be on map
	}
	else
	{	// key is positive
		y &= 31;								// y is implicitly on the map
	}

	// generate initial x
	Uint32 x;

	x = key % 31;								// x is implicitly on the map

	// Prepare loop
	Uint32 cnt, diff, okey;

	cnt = ((1024 + 31) - y) >> 5;				// count of available iterations before end of map
	diff = 1024 - x;							// diff is a number between 994 and 1024
	okey = key;									// remember what the key used to be

	// Loop until counter == 0
	Uint32 i;
	BYTE tile;

	while (cnt--)
	{
		for (i = 0; i < diff; i += 31)	// account for the remainder of a line of tiles
		{
			tile = mapData[getLinear(x + i, y)];

			if (tile)
			{	// speedup: most tiles are blank
				if ((tile < vieNormalEnd) || (tile == vieSafeZone))
				{	// basic tiles only
					if (tile == vieBorder)
					{
						if (x == 0)		continue;
						if (y == 0)		continue;
						if (x == 1023)	continue;
						if (y == 1023)	continue;
					}

					key += okey ^ tile;
				}
			}
		}

		y += 32;								// skip ahead 32 lines
	}

	return key;
}

Uint32 generateParameterChecksum(Uint32 key, Uint32 * settings)
{
	Uint32 i, c = 0;

	for (i = 0; i < 357; i++)
	{
		c += (settings[i] ^ key);
	}

	return c;
}

Uint32 generatePartialEXE(Uint32 key, char * buffer, Uint32 len)
{
	Uint32 data,
		   partialKey = 0,
		   offset = 0;

	while (len > 4)
	{
		data = getLong(buffer, offset);

		if ((len & 0x7FF) == 0)
		{
			partialKey += (data | key);
		}
		else
		{
			partialKey ^= (data ^ key);
		}

		len -= 3;
		offset += 3;
	}

	delete []buffer;

	return partialKey;
}

Uint32 generateEXEChecksum(Uint32 key, char * fileName)
{
	fileAccess fa;
	char * fd;

	if (fa.open(fileName, true))
	{
		Uint32 c = 0;

		c += generatePartialEXE(key, fd = fa.readStuff(0x00006940, 0x0000D5A0), 0x0000D5A0);
		c += generatePartialEXE(key, fd = fa.readStuff(0x00000400, 0x00002480), 0x00002480);
		c += generatePartialEXE(key, fd = fa.readStuff(0x0001CAE0, 0x00006F70), 0x00006F70);
		c += generatePartialEXE(key, fd = fa.readStuff(0x00042E70, 0x000021E0), 0x000021E0);
		c += generatePartialEXE(key, fd = fa.readStuff(0x000305A0, 0x0000D3C0), 0x0000D3C0);
		c += generatePartialEXE(key, fd = fa.readStuff(0x0002A5B0, 0x00005FE0), 0x00005FE0);
		c += generatePartialEXE(key, fd = fa.readStuff(0x00003960, 0x00001C90), 0x00001C90);
		c += generatePartialEXE(key, fd = fa.readStuff(0x00013EF0, 0x00008BE0), 0x00008BE0);
		c += generatePartialEXE(key, fd = fa.readStuff(0x00005750, 0x000000E0), 0x000000E0);

		return c;
	}
	else
	{
		return -1;
	}
}
